/***************************************************************************
                           PRINTER.CPP
                        -------------------                          
    begin                : 2001-03-09
    copyright            : (C) 2000 by Jan Becker
    email                : tahoma@gmx.de
 	author               : Jan Becker
                                                                          
                                                                          
     This program is free software; you can redistribute it and/or modify  
     it under the terms of the GNU General Public License as published by  
     the Free Software Foundation; either version 2 of the License, or     
     (at your option) any later version.                                   
                                                                          
 ***************************************************************************/

#include "stdafx.h"
#include "printer.h"

#include <windows.h>
#include <winspool.h>


LPMONITOR_INFO_1 m_lpMonitorInfo1;
LPDRIVER_INFO_1 m_lpDriverInfo1;
LPPRINTER_INFO_1 m_lpPrinterInfo1;  
PRINTER_DEFAULTS m_pPrinterDef;
MONITOR_INFO_2 m_MonitorInfo2;
DRIVER_INFO_2 m_DriverInfo2;
PRINTER_INFO_2 m_PrinterInfo2;

/********************************************************************/
/***   GetPrinterDirectory: Findet den Druckerinstallationspfad   ***/
/********************************************************************/
int WINAPI GetPrinterDirectory(LPSTR string)
{	
	DWORD dwNumber;
		
	if ( !GetPrinterDriverDirectory(NULL, NULL, 1, (LPBYTE)string, 255, &dwNumber) )
		return PRINTER_ERR_API;	

	return 1;
}


/********************************************************************/
/***   InstallASPrinter: Installiert den AS - Drucker             ***/
/********************************************************************/
int WINAPI InstallMyPrinter(LPSTR lpMonitor, LPSTR lpMonitorFile,
							LPSTR lpNamePrinter, LPSTR lpNamePrinterDriver,
							LPSTR lpDriverFile, LPSTR lpDataFile,
							LPSTR lpConfigFile, LPSTR lpNamePort) {
  int iRet;  
    
  iRet = IsMonitorInstalled(lpMonitor);
  if ( iRet == PRINTER_ERR_API ) return PRINTER_ERR_MONITORCHECK;
  
  if ( !iRet ) {
    if ( !InstallMonitor(lpMonitor, lpMonitorFile) ) 
		return PRINTER_ERR_MONITOR_INSTALL_FAILED;
  }
 
  iRet = IsPrinterDriverInstalled(lpNamePrinterDriver);
  if ( iRet == PRINTER_ERR_API ) 
	  return PRINTER_ERR_DRIVERCHECK;
  
  if ( !iRet ) {
	if( !InstallPrinterDriver(lpNamePrinterDriver, lpDriverFile, lpDataFile, lpConfigFile) )
		return PRINTER_ERR_DRIVER_INSTALL_FAILED; 
  } 

    
  iRet = IsPrinterInstalled(lpNamePrinter);
  if ( iRet == PRINTER_ERR_API ) 
		return PRINTER_ERR_PRINTERCHECK;
  
  if ( iRet == PRINTER_ERR_API_REMOVE )
		return PRINTER_ERR_UPDATE_FAILED;

  if ( !iRet ) {
	  if (!InstallPrinter(lpNamePrinter, lpNamePort,
                          lpNamePrinterDriver))  
		return PRINTER_ERR_PRINTER_INSTALL_FAILED;
  }  

  return 1;
}


/********************************************************************/
/***   InstallASPrinter: Installiert den AS - Drucker unter 95/98 ***/
/********************************************************************/
int WINAPI InstallMyPrinter9x(LPSTR lpMonitor, LPSTR lpMonitorFile,
							  LPSTR lpNamePrinter, LPSTR lpNamePrinterDriver,
							  LPSTR lpDriverFile, LPSTR lpDataFile,
							  LPSTR lpConfigFile, LPSTR lpNamePort) {
  int iRet;      

  iRet = IsPrinterDriverInstalled(lpNamePrinterDriver);
  if ( iRet == PRINTER_ERR_API ) return PRINTER_ERR_DRIVERCHECK;
  
  if ( !iRet ) {
	if ( !InstallPrinterDriver(lpNamePrinterDriver, lpDriverFile, lpDataFile, lpConfigFile, 0))
        return PRINTER_ERR_DRIVER_INSTALL_FAILED; 
  }

  iRet = IsPrinterInstalled(lpNamePrinter);
  if ( iRet == PRINTER_ERR_API ) 
	  return PRINTER_ERR_PRINTERCHECK;
  if ( iRet == PRINTER_ERR_API_REMOVE ) 
	  return PRINTER_ERR_UPDATE_FAILED;
  
  if ( !iRet ) {	  
	  WritePrivateProfileString("Ports", lpNamePort, "", "WIN.INI");
	  WritePrivateProfileString(NULL, NULL, NULL, "WIN.INI"); //flush for w95

	  if ( !InstallPrinter(lpNamePrinter, lpNamePort, lpNamePrinterDriver, FALSE) )
        return PRINTER_ERR_PRINTER_INSTALL_FAILED; 
  }

  return 1; 
}


/********************************************************************/
/***   IsMonitorInstalled: Prft, ob Monitor installiert ist      ***/
/********************************************************************/
int WINAPI IsMonitorInstalled(LPSTR lpMonitorName, BOOL bDeleteIfFound)
{
  HANDLE hBuffer;
  LPBYTE lpBuffer = NULL;
  DWORD  dwNeeded = 0;
  DWORD  dwReturned = 0;
  DWORD  dwIndex = 0;
  BOOL   bInstalled = 0;
  
  // Ermittle Puffergre
  EnumMonitors(NULL, 1, NULL, 0, &dwNeeded, &dwReturned); 

  // Fordere Puffer an
  hBuffer = GlobalAlloc(GHND, dwNeeded);  
  if ( hBuffer == NULL ) {	  
	  return PRINTER_ERR_MEMORY;
  }
  
  lpBuffer = (LPBYTE) GlobalLock(hBuffer);
  if ( lpBuffer == NULL ) {
	  return PRINTER_ERR_MEMORY;
  }
  
  // Ermittle alle Monitore
  bInstalled = FALSE;
  if ( EnumMonitors(NULL, 1,  lpBuffer, dwNeeded, &dwNeeded,
                   &dwReturned) ) {
     m_lpMonitorInfo1 = (LPMONITOR_INFO_1)lpBuffer;
	 if ( m_lpMonitorInfo1 != NULL ) {
		 for ( dwIndex = 0; dwIndex < dwReturned; dwIndex++ ) {
			 if( m_lpMonitorInfo1[dwIndex].pName != NULL ) {
			   if ( !lstrcmpi(m_lpMonitorInfo1[dwIndex].pName, lpMonitorName) )
				 bInstalled = TRUE;			   
			 }
			 else
				 // monitorbuffer invalid (==NULL)		 
				 return PRINTER_ERR_API;
		 }
	 }
  }
  else
	  return PRINTER_ERR_API;
  
  GlobalUnlock(hBuffer);
  GlobalFree(hBuffer);

  if ( bInstalled && bDeleteIfFound )
	  if ( !DeleteMonitor(NULL, NULL, lpMonitorName) )
		  return PRINTER_ERR_API_REMOVE;
	  else
		  bInstalled = FALSE;

  return bInstalled?1:0;
}

/********************************************************************/
/***   IsPrinterDriverInstalled: Prft, ob Druckertreiber         ***/
/***                             installiert ist                  ***/
/********************************************************************/
int WINAPI IsPrinterDriverInstalled(LPSTR lpPrinterDriverName, BOOL bDeleteIfFound)
{
  HANDLE hBuffer = NULL;
  LPBYTE lpBuffer = NULL;
  DWORD  dwNeeded = 0;
  DWORD  dwReturned = 0;
  DWORD  dwIndex = 0;
  BOOL   bInstalled = FALSE;

  EnumPrinterDrivers(NULL, NULL, 1, NULL, 0, &dwNeeded, &dwReturned);

  // Fordere Puffer an
  if ( dwNeeded > 0 )
	hBuffer = GlobalAlloc(GHND, dwNeeded);  
  else
    return 0;
  
  if ( hBuffer == NULL ) {
	  // konnte speicher nicht anfordern
	  return PRINTER_ERR_MEMORY;
  }//if
  
  lpBuffer = (LPBYTE) GlobalLock(hBuffer);
  if ( lpBuffer == NULL ) {
	  // mal wieder kein speicher da
	  return PRINTER_ERR_MEMORY;
  }
  
  // Ermittle alle Druckertreiber
  bInstalled = FALSE;
  if (EnumPrinterDrivers(NULL, NULL, 1,  lpBuffer, dwNeeded, &dwNeeded, &dwReturned)) {
     m_lpDriverInfo1 = (LPDRIVER_INFO_1)lpBuffer;
	 if( m_lpDriverInfo1 != NULL ) {
		 for ( dwIndex = 0; dwIndex < dwReturned; dwIndex++ ) {
			 if( m_lpDriverInfo1[dwIndex].pName != NULL ) {
				if ( !lstrcmpi(m_lpDriverInfo1[dwIndex].pName, lpPrinterDriverName) )				 
				   bInstalled = TRUE;				 
			 }
			 else
				// driverbuffer invalid (==NULL)
				return PRINTER_ERR_API;
		 }
	 }
  }
  else  	  
	  // fehler beim ausfuehren von EnumPrinterDrivers
	  return PRINTER_ERR_API;

  GlobalUnlock(hBuffer);
  GlobalFree(hBuffer);
  
  if ( bInstalled && bDeleteIfFound )
	  if ( !DeletePrinterDriver(NULL, NULL, lpPrinterDriverName) )
		  return PRINTER_ERR_API_REMOVE;
	  else
		  bInstalled = FALSE;
  
  return bInstalled?1:0;
}

/********************************************************************/
/***   IsPrinterInstalled: Prft, ob Drucker installiert ist      ***/
/********************************************************************/
int WINAPI IsPrinterInstalled(LPSTR lpPrinterName, BOOL bDeleteIfFound) {
  HANDLE hBuffer;
  HANDLE hPrinter;
  LPBYTE lpBuffer = NULL;
  DWORD  dwNeeded = 0;
  DWORD  dwReturned = 0;
  DWORD  dwIndex = 0;
  BOOL   bInstalled = 0;
  DWORD dwError = 0;

  EnumPrinters(PRINTER_ENUM_LOCAL, NULL, 1, NULL, 0, &dwNeeded, &dwReturned);
  if ( dwNeeded == 0 )
	  return 0;

  hBuffer = GlobalAlloc(GHND, dwNeeded);  
  if ( hBuffer == NULL )
  	  // konnte keinen speicher kriegen
	  return PRINTER_ERR_MEMORY;

  lpBuffer = (LPBYTE)GlobalLock(hBuffer);
  if ( lpBuffer == NULL ) {
	  // speicher konnte nicht gelockt werden
	  return PRINTER_ERR_MEMORY;
  }
  
  // Ermittle alle Drucker
  bInstalled = FALSE;
  if ( EnumPrinters(PRINTER_ENUM_LOCAL, NULL, 1,  lpBuffer, dwNeeded, &dwNeeded, &dwReturned) ) {    
     m_lpPrinterInfo1 = (LPPRINTER_INFO_1)lpBuffer;
	 if( m_lpPrinterInfo1 != NULL ) {
		 for ( dwIndex = 0; dwIndex < dwReturned; dwIndex++ ) {
			 if ( m_lpPrinterInfo1[dwIndex].pName != NULL ) {
			   if ( !lstrcmpi(m_lpPrinterInfo1[dwIndex].pName, lpPrinterName) )			   
				 bInstalled = TRUE;			   
			 }
			 else
				// buffer invalid (==NULL)
				return PRINTER_ERR_API;
		 }
	 }
  }
  else  	  
	  return PRINTER_ERR_API;
 
  GlobalUnlock(hBuffer);
  GlobalFree(hBuffer);  

  if ( bDeleteIfFound && bInstalled ) {
  
	  ZeroMemory(&m_pPrinterDef, sizeof(PRINTER_DEFAULTS));

	  m_pPrinterDef.DesiredAccess = PRINTER_ALL_ACCESS;

	  if ( OpenPrinter(lpPrinterName, &hPrinter, &m_pPrinterDef) ) {
		 if ( DeletePrinter(hPrinter) )
			 bInstalled = FALSE;
		 else		 
			return PRINTER_ERR_API_REMOVE;
	  }
	  else	  			
		 // openprinter failed	
		 return PRINTER_ERR_API_OPEN;
  }
  
      
  return bInstalled?1:0;
}


/********************************************************************/
/***   InstallMonitor: Installiert einen Print Monitor            ***/
/********************************************************************/
int WINAPI InstallMonitor(LPSTR lpMonitorName, LPSTR lpDllName) {
  // Flle Struktur auf
  m_MonitorInfo2.pName        = lpMonitorName;
#ifdef WINDOWS95  
  m_MonitorInfo2.pEnvironment = "Windows 4.0"; //NULL;
#else
  m_MonitorInfo2.pEnvironment = NULL;
#endif
  m_MonitorInfo2.pDLLName     = lpDllName;

  if ( !AddMonitor(NULL, 2, (LPBYTE) &m_MonitorInfo2) )
	return PRINTER_ERR_API;
  
  return 1;
}

/********************************************************************/
/***   InstallPrinterDriver: Installiert einen Druckertreiber     ***/
/********************************************************************/
int WINAPI InstallPrinterDriver(LPSTR lpPrinterDriverName, LPSTR lpDriverFile,
                          LPSTR lpDataFile, LPSTR lpConfigFile, DWORD cVersion) {

  ZeroMemory(&m_DriverInfo2, sizeof(DRIVER_INFO_2));

  // Flle Struktur auf
  m_DriverInfo2.cVersion     = cVersion;
  m_DriverInfo2.pName        = lpPrinterDriverName;
  m_DriverInfo2.pEnvironment = NULL;
  m_DriverInfo2.pDriverPath  = lpDriverFile;
  m_DriverInfo2.pDataFile    = lpDataFile;
  m_DriverInfo2.pConfigFile  = lpConfigFile;
  
  if( !AddPrinterDriver(NULL, 2, (LPBYTE) &m_DriverInfo2) )
  	  // ging schief
	  return PRINTER_ERR_API;

  return 1;
}


int WINAPI PrinterPortExists(LPSTR lpPortName, BOOL bDeleteIfFound) {
	DWORD dwNeeded = 0, dwPortCount = 0;
	HANDLE hBuffer = NULL;
	LPBYTE lpBuffer = NULL;
	LPPORT_INFO_1 lpInfo = NULL;

	EnumPorts(NULL, 1, NULL, 0, &dwNeeded, &dwPortCount);
	if ( dwNeeded > 0 ) {
		hBuffer = GlobalAlloc(GHND, dwNeeded);
		if ( hBuffer == NULL ) {
			return PRINTER_ERR_MEMORY;
		}
		lpBuffer = (LPBYTE) GlobalLock(hBuffer);
		if ( lpBuffer == NULL ) {			
			GlobalFree(hBuffer);
			return PRINTER_ERR_MEMORY;
		}
		if ( !EnumPorts(NULL, 1, lpBuffer, dwNeeded, &dwNeeded, &dwPortCount) ) {
			GlobalUnlock(hBuffer);
			GlobalFree(hBuffer);
			return PRINTER_ERR_API;
		}
	}
	else
		return PRINTER_ERR_API;

	lpInfo = (LPPORT_INFO_1)lpBuffer;
	for ( DWORD i = 0; i < dwPortCount; i++ ) {
		if ( lstrcmpi(lpInfo[i].pName, lpPortName) == 0 ) {
			GlobalUnlock(hBuffer);
			GlobalFree(hBuffer);

			if ( bDeleteIfFound )
				return DeletePort(NULL, NULL, lpPortName)?PRINTER_ERR_API_REMOVE:0;
				
			return 1;
		}
	}
	
	GlobalUnlock(hBuffer);
	GlobalFree(hBuffer);	

	return 0;
}

/********************************************************************/
/***   InstallPrinter: Installiert einen Drucker                  ***/
/********************************************************************/
int WINAPI InstallPrinter(LPSTR lpPrinterName, LPSTR lpPortName,
                    LPSTR lpDriverName, BOOL bConfigDialog, HWND hWnd) {
  DWORD			 dwError;
  HANDLE         hPrinter = NULL;
  BOOL           bConfig;
  PRINTER_INFO_2* pInfo2;

  // eigentlich muss das nur als dummy gemacht werden.
  // komischerweise erkennt winnt einen neuen port nicht,
  // wenn man die portliste nicht abgerufen hat oder wenigstens
  // einmal im Drucker/Datei/Servereigenschaften/Anschlsse
  // war...
  if ( !PrinterPortExists(lpPortName) )
	// port wurde nicht gefunden
	return PRINTER_ERR_API;
  
  // Flle Struktur auf
  memset(&m_PrinterInfo2, 0, sizeof(PRINTER_INFO_2));
  m_PrinterInfo2.pServerName     = NULL;
  m_PrinterInfo2.pPrinterName    = lpPrinterName;
  m_PrinterInfo2.pPortName       = lpPortName;
  m_PrinterInfo2.pDriverName     = lpDriverName;
  m_PrinterInfo2.pPrintProcessor = NAME_WINPRINT;
  m_PrinterInfo2.pDatatype       = NAME_RAW;
  
  // Installiere Drucker
  hPrinter = AddPrinter(NULL, 2, (LPBYTE) &m_PrinterInfo2);
  
  if ( hPrinter == NULL )
	// addprinter failed
	return PRINTER_ERR_API;
  
  // soll nicht gleich konfiguriert werden, dann raus
  if ( bConfigDialog == FALSE ) {
	ClosePrinter(hPrinter);    
	return 1;
  }
  
  DWORD dwNeeded = 0;
  DWORD dwNeed1 = 0;
  HANDLE hBuffer = NULL;
  LPBYTE lpBuffer = NULL;
  
  GetPrinter(hPrinter, 2, NULL, 0, &dwNeeded);
  if ( dwNeeded > 0 )
  {
	  hBuffer = GlobalAlloc(GHND, dwNeeded);  
	  if ( hBuffer == NULL ) return PRINTER_ERR_MEMORY;

	  lpBuffer = (LPBYTE) GlobalLock(hBuffer);
	  if ( hBuffer == NULL ) return PRINTER_ERR_MEMORY;

	  if ( GetPrinter(hPrinter, 2, lpBuffer, dwNeeded, &dwNeed1) ) {
		  
		  pInfo2 = (LPPRINTER_INFO_2)lpBuffer;	  
		  dwNeeded = (DWORD)DocumentProperties(hWnd, hPrinter, (LPSTR)pInfo2->pPrinterName, NULL, NULL, 0);
		  
		  if ( dwNeeded > 0 ) {
			
			  if ( DocumentProperties(hWnd, hPrinter, (LPSTR)pInfo2->pPrinterName, pInfo2->pDevMode, pInfo2->pDevMode, DM_COPY|DM_MODIFY|DM_PROMPT) < 0 )
				bConfig = FALSE;
				
			  if ( !SetPrinter(hPrinter, 2,(LPBYTE)pInfo2,0) ) {
				dwError = GetLastError();
				if ( (dwError != 5) && (dwError != 0) )						
					bConfig = FALSE;					
			  }
		  }
		  else 
			  bConfig = FALSE;
	  }
	  else
		  // getprinter failed
		  bConfig = FALSE;
  }
  
  
  ClosePrinter(hPrinter);
  GlobalUnlock(hBuffer);
  GlobalFree(hBuffer);
  
  return bConfig?1:0;
}
